#ifndef ComInterpNodes_H
#define ComInterpNodes_H

#include "ComBasic3D.h"


/** communicate ghost cells and sum the contribution with a index indicating the number of species*/
inline void communicateInterp(int nx, int ny, int nz, int ns, double ****vector, int bcFaceXright, int bcFaceXleft, int bcFaceYright, int bcFaceYleft,int bcFaceZright,int bcFaceZleft, VirtualTopology3D *vct){
   // allocate 6 ghost cell Faces
  double *ghostXrightFace = new double[(ny-2)*(nz-2)];
  double *ghostXleftFace  = new double[(ny-2)*(nz-2)];
  double *ghostYrightFace = new double[(nx-2)*(nz-2)];
  double *ghostYleftFace  = new double[(nx-2)*(nz-2)];
  double *ghostZrightFace = new double[(nx-2)*(ny-2)];
  double *ghostZleftFace  = new double[(nx-2)*(ny-2)];
  // allocate 12 ghost cell Edges
  // X EDGE
  double *ghostXsameYleftZleftEdge   = new double[nx-2];
  double *ghostXsameYrightZleftEdge  = new double[nx-2];
  double *ghostXsameYleftZrightEdge  = new double[nx-2];
  double *ghostXsameYrightZrightEdge = new double[nx-2];
  // Y EDGE
  double *ghostXrightYsameZleftEdge  = new double[ny-2];
  double *ghostXleftYsameZleftEdge   = new double[ny-2];
  double *ghostXrightYsameZrightEdge = new double[ny-2];
  double *ghostXleftYsameZrightEdge  = new double[ny-2];
  // Z EDGE
  double *ghostXrightYleftZsameEdge  = new double[nz-2];
  double *ghostXrightYrightZsameEdge = new double[nz-2];
  double *ghostXleftYleftZsameEdge   = new double[nz-2];
  double *ghostXleftYrightZsameEdge  = new double[nz-2];
  // allocate 8 ghost cell corner
  double ghostXrightYrightZrightCorner,ghostXleftYrightZrightCorner,ghostXrightYleftZrightCorner,ghostXleftYleftZrightCorner;
  double ghostXrightYrightZleftCorner, ghostXleftYrightZleftCorner, ghostXrightYleftZleftCorner, ghostXleftYleftZleftCorner;
  
  // apply boundary condition to 6 Ghost Faces and communicate if necessary to 6 processors: along 3 DIRECTIONS
  makeCenterFace(nx,ny,nz,vector,ns,ghostXrightFace,ghostXleftFace,ghostYrightFace,ghostYleftFace,ghostZrightFace,ghostZleftFace);
   //communication
  communicateGhostFace((ny-2)*(nz-2),vct->getCartesian_rank(),vct->getXright_neighbor_P(),vct->getXleft_neighbor_P(),0,vct->getXLEN(),vct->getYLEN(),vct->getZLEN(),ghostXrightFace, ghostXleftFace);
  communicateGhostFace((nx-2)*(nz-2),vct->getCartesian_rank(),vct->getYright_neighbor_P(),vct->getYleft_neighbor_P(),1,vct->getXLEN(),vct->getYLEN(),vct->getZLEN(),ghostYrightFace, ghostYleftFace);
  communicateGhostFace((nx-2)*(ny-2),vct->getCartesian_rank(),vct->getZright_neighbor_P(),vct->getZleft_neighbor_P(),2,vct->getXLEN(),vct->getYLEN(),vct->getZLEN(),ghostZrightFace, ghostZleftFace);
  addFace(nx,ny,nz,vector,ns,ghostXrightFace,ghostXleftFace,ghostYrightFace,ghostYleftFace,ghostZrightFace,ghostZleftFace,vct);
   /** prepare ghost cell Edge Y for communication: these are communicate: these are communicated in X direction */
  makeNodeEdgeY(nx,ny,nz,ghostZleftFace,ghostZrightFace,ghostXrightYsameZrightEdge,ghostXleftYsameZleftEdge,ghostXleftYsameZrightEdge,ghostXrightYsameZleftEdge);
  /** prepare ghost cell Edge Z for communication: these are communicated in Y direction */
  makeNodeEdgeZ(nx,ny,nz,ghostXleftFace,ghostXrightFace,ghostXrightYrightZsameEdge,ghostXleftYleftZsameEdge,ghostXrightYleftZsameEdge,ghostXleftYrightZsameEdge);
  /** prepare ghost cell Edge X for communication: these are communicated in  Z direction*/
  makeNodeEdgeX(nx,ny,nz,ghostYleftFace,ghostYrightFace,ghostXsameYrightZrightEdge,ghostXsameYleftZleftEdge,ghostXsameYleftZrightEdge,ghostXsameYrightZleftEdge);
  
  // communicate twice each direction
  // X-DIRECTION: Z -> X -> Y
  //;
  communicateGhostFace(ny-2,vct->getCartesian_rank(),vct->getXright_neighbor_P(),vct->getXleft_neighbor_P(),0,vct->getXLEN(),vct->getYLEN(),vct->getZLEN(),ghostXrightYsameZleftEdge,ghostXleftYsameZleftEdge);
  communicateGhostFace(ny-2,vct->getCartesian_rank(),vct->getXright_neighbor_P(),vct->getXleft_neighbor_P(),0,vct->getXLEN(),vct->getYLEN(),vct->getZLEN(),ghostXrightYsameZrightEdge,ghostXleftYsameZrightEdge);
  // Y-DIRECTION: X -> Y -> Z
  //;
  communicateGhostFace(nz-2,vct->getCartesian_rank(),vct->getYright_neighbor_P(),vct->getYleft_neighbor_P(),1,vct->getXLEN(),vct->getYLEN(),vct->getZLEN(),ghostXleftYrightZsameEdge,ghostXleftYleftZsameEdge);
  communicateGhostFace(nz-2,vct->getCartesian_rank(),vct->getYright_neighbor_P(),vct->getYleft_neighbor_P(),1,vct->getXLEN(),vct->getYLEN(),vct->getZLEN(),ghostXrightYrightZsameEdge,ghostXrightYleftZsameEdge);
  // Z-DIRECTION: Y -> Z
  //;
  communicateGhostFace(nx-2,vct->getCartesian_rank(),vct->getZright_neighbor_P(),vct->getZleft_neighbor_P(),2,vct->getXLEN(),vct->getYLEN(),vct->getZLEN(),ghostXsameYleftZrightEdge,ghostXsameYleftZleftEdge);
  communicateGhostFace(nx-2,vct->getCartesian_rank(),vct->getZright_neighbor_P(),vct->getZleft_neighbor_P(),2,vct->getXLEN(),vct->getYLEN(),vct->getZLEN(),ghostXsameYrightZrightEdge,ghostXsameYrightZleftEdge);
  //parse
  //;
  addEdgeZ(nx,ny,nz,vector,ns,ghostXrightYrightZsameEdge,ghostXleftYleftZsameEdge,ghostXrightYleftZsameEdge,ghostXleftYrightZsameEdge,vct);
  addEdgeY(nx,ny,nz,vector,ns,ghostXrightYsameZrightEdge,ghostXleftYsameZleftEdge,ghostXleftYsameZrightEdge,ghostXrightYsameZleftEdge,vct);
  addEdgeX(nx,ny,nz,vector,ns,ghostXsameYrightZrightEdge,ghostXsameYleftZleftEdge,ghostXsameYleftZrightEdge,ghostXsameYrightZleftEdge,vct);
                              
	 
 
  makeNodeCorner(nx,ny,nz,ghostXsameYrightZrightEdge,ghostXsameYleftZleftEdge,ghostXsameYleftZrightEdge,ghostXsameYrightZleftEdge,&ghostXrightYrightZrightCorner,&ghostXleftYrightZrightCorner,&ghostXrightYleftZrightCorner,&ghostXleftYleftZrightCorner,&ghostXrightYrightZleftCorner,&ghostXleftYrightZleftCorner,&ghostXrightYleftZleftCorner,&ghostXleftYleftZleftCorner);
 // communicate only in the X-DIRECTION
  communicateGhostFace(1,vct->getCartesian_rank(),vct->getXright_neighbor_P(),vct->getXleft_neighbor_P(),0,vct->getXLEN(),vct->getYLEN(),vct->getZLEN(),&ghostXrightYrightZrightCorner,&ghostXleftYrightZrightCorner);
  communicateGhostFace(1,vct->getCartesian_rank(),vct->getXright_neighbor_P(),vct->getXleft_neighbor_P(),0,vct->getXLEN(),vct->getYLEN(),vct->getZLEN(),&ghostXrightYleftZrightCorner,&ghostXleftYleftZrightCorner);
  communicateGhostFace(1,vct->getCartesian_rank(),vct->getXright_neighbor_P(),vct->getXleft_neighbor_P(),0,vct->getXLEN(),vct->getYLEN(),vct->getZLEN(),&ghostXrightYleftZleftCorner,&ghostXleftYleftZleftCorner);
  communicateGhostFace(1,vct->getCartesian_rank(),vct->getXright_neighbor_P(),vct->getXleft_neighbor_P(),0,vct->getXLEN(),vct->getYLEN(),vct->getZLEN(),&ghostXrightYrightZleftCorner,&ghostXleftYrightZleftCorner);
  // parse
  addCorner(nx,ny,nz,vector,ns,&ghostXrightYrightZrightCorner,&ghostXleftYrightZrightCorner,&ghostXrightYleftZrightCorner,&ghostXleftYleftZrightCorner,&ghostXrightYrightZleftCorner,&ghostXleftYrightZleftCorner,&ghostXrightYleftZleftCorner,&ghostXleftYleftZleftCorner,vct);
 

  delete[] ghostXrightFace;
  delete[] ghostXleftFace;
  delete[] ghostYrightFace;
  delete[] ghostYleftFace;
  delete[] ghostZrightFace;
  delete[] ghostZleftFace;
  // X EDGE
  delete[] ghostXsameYleftZleftEdge;
  delete[]ghostXsameYrightZleftEdge;
  delete[] ghostXsameYleftZrightEdge;
  delete[] ghostXsameYrightZrightEdge;
  // Y EDGE
  delete[] ghostXrightYsameZleftEdge;
  delete[] ghostXleftYsameZleftEdge;
  delete[] ghostXrightYsameZrightEdge;
  delete[] ghostXleftYsameZrightEdge;
  // Z EDGE
  delete[] ghostXrightYleftZsameEdge;
  delete[] ghostXrightYrightZsameEdge;
  delete[] ghostXleftYleftZsameEdge;
  delete[] ghostXleftYrightZsameEdge;
}













#endif
