#ifndef Field_H
#define Field_H

#include <vector>

class Field {
public:
  /** print field information */
  virtual void print(void) const = 0;


  /** get Potential array */
  virtual double ***getPHI() = 0;
  /** get Electric field X component array */
  virtual double ***getEx() = 0;
  /** get Electric field Y component array */
  virtual double ***getEy() = 0;
  /** get Electric field Z component array */
  virtual double ***getEz() = 0;
  /** get Magnetic field X component array */
  virtual double ***getBx() = 0;
  /** get Magnetic field Y component array */
  virtual double ***getBy() = 0;
  /** get Magnetic field Z component array */
  virtual double ***getBz() = 0;
  /** get Electric Field component X defined on node(indexX,indexY,indexZ) */
  virtual double &getEx(int indexX, int indexY, int indexZ) const = 0;
  /** get Electric Field component Y defined on node(indexX,indexY,indexZ) */
  virtual double &getEy(int indexX, int indexY, int indexZ) const = 0;
  /** get Electric Field component Z defined on node(indexX,indexY,indexZ) */
  virtual double &getEz(int indexX, int indexY, int indexZ) const = 0;
  /** get Magnetic Field component X defined on node(indexX,indexY,indexZ) */
  virtual double &getBx(int indexX, int indexY, int indexZ) const = 0;
  /** get Magnetic Field component Y defined on node(indexX,indexY,indexZ) */
  virtual double &getBy(int indexX, int indexY, int indexZ) const = 0;
  /** get Magnetic Field component Z defined on node(indexX,indexY,indexZ) */
  virtual double &getBz(int indexX, int indexY, int indexZ) const = 0;
  /** get density on cell(indexX,indexY,indexZ)  */
  virtual double &getRHOc(int indexX, int indexY, int indexZ) const = 0;
  /** get density on node(indexX,indexY,indexZ)  */
  virtual double &getRHOn(int indexX, int indexY, int indexZ) const = 0;
  /** SPECIES: get density defined on center of cells*/
  virtual double &getRHOcs(int indexX, int indexY, int indexZ, int ns) const = 0;
  /** SPECIES: get density defined on nodes */
  virtual double &getRHOns(int indexX, int indexY, int indexZ, int ns) const = 0;
  /** get current -Direction X */
  virtual double &getJx(int indexX, int indexY, int indexZ) const = 0;
  /** get current -Direction Y  */
  virtual double &getJy(int indexX, int indexY, int indexZ) const = 0;
  /** get current -Direction Z */
  virtual double &getJz(int indexX, int indexY, int indexZ) const = 0;
  virtual double &getJxs(int indexX, int indexY, int indexZ, int is) const = 0;
  /** get current -Direction Y  */
  virtual double &getJys(int indexX, int indexY, int indexZ, int is) const = 0;
  /** get current -Direction Z */
  virtual double &getJzs(int indexX, int indexY, int indexZ, int is) const = 0;

  /** get density array defined on centers cells */
  virtual double ***getRHOc() = 0;
  /** get density array defined on nodes*/
  virtual double ***getRHOn() = 0;
  /** SPECIES: get density array defined on nodes  */
  virtual double ****getRHOns() = 0;
  /** get current array X component */
  virtual double ***getJx() = 0;
  /** get current array Y component */
  virtual double ***getJy() = 0;
  /** get current array Z component */
  virtual double ***getJz() = 0;
  /** SPECIES: get current array X component */
  virtual double ****getJxs() = 0;
  /** SPECIES: get current array Y component */
  virtual double ****getJys() = 0;
  /** SPECIES: get current array X component */
  virtual double ****getJzs() = 0;
  /** SPECIES: get pressure tensor component XX defined on nodes */
  virtual double ****getpXXsn() = 0;
  /** SPECIES: get pressure tensor component XY defined on nodes */
  virtual double ****getpXYsn() = 0;
  /** SPECIES: get pressure tensor component XZ defined on nodes */
  virtual double ****getpXZsn() = 0;
  /** SPECIES: get pressure tensor component YY defined on nodes */
  virtual double ****getpYYsn() = 0;
  /** SPECIES: get pressure tensor component YZ defined on nodes */
  virtual double ****getpYZsn() = 0;
  /** SPECIES: get pressure tensor component ZZ defined on nodes */
  virtual double ****getpZZsn() = 0;

    //Ohms Ohm's law
    virtual double ****getRHOns_collect() = 0;
    virtual double ****getJxs_collect() = 0;
    virtual double ****getJys_collect() = 0;
    virtual double ****getJzs_collect() = 0;
    virtual double ****getpXZsn_collect() = 0;
    virtual double ****getpYZsn_collect() = 0;
    virtual double getRHOns_combined_vector(int indexX, int indexY, int indexZ, std::vector<int> ns, double coeff) const = 0;

    virtual double getJxs_combined_vector(int indexX, int indexY, int indexZ, std::vector<int> ns, double coeff) const = 0;
    virtual double getJys_combined_vector(int indexX, int indexY, int indexZ, std::vector<int> ns, double coeff) const = 0;
    virtual double getJzs_combined_vector(int indexX, int indexY, int indexZ, std::vector<int> ns, double coeff) const = 0;

    virtual double getB0x() const = 0;
    virtual double getB0y() const = 0;
    virtual double getB0z() const = 0;

    virtual double &getRHOns_collect(int indexX, int indexY, int indexZ, int is) const = 0;
    virtual double &getJxs_collect(int indexX, int indexY, int indexZ, int is) const = 0;
    virtual double &getJys_collect(int indexX, int indexY, int indexZ, int is) const = 0;
    virtual double &getJzs_collect(int indexX, int indexY, int indexZ, int is) const = 0;
    virtual double &getpXZsn_collect(int indexX, int indexY, int indexZ, int is) const = 0;
    virtual double &getpYZsn_collect(int indexX, int indexY, int indexZ, int is) const = 0;



  // //////////////////////// INTERPOLATION ///////////////////////////////
  /** set to 0 all the densities fields */
  virtual void setZeroDensities() = 0;
  /** add an amount of charge density to charge density field at node X,Y,Z */
  virtual void addRho(double weight[][2][2], int X, int Y, int Z, int ns) = 0;
  /** add an amount of current density - direction X to current density field at node X,Y,Z */
  virtual void addJx(double weight[][2][2], int X, int Y, int Z, int ns) = 0;
  /** add an amount of current density - direction Y to current density field at node X,Y,Z */
  virtual void addJy(double weight[][2][2], int X, int Y, int Z, int ns) = 0;
  /** add an amount of current density - direction Z to current density field at node X,Y,Z */
  virtual void addJz(double weight[][2][2], int X, int Y, int Z, int ns) = 0;
  /** add an amount of pressure density - direction XX to current density field at node X,Y,Z */
  virtual void addPxx(double weight[][2][2], int X, int Y, int Z, int ns) = 0;
  /** add an amount of pressure density - direction XY to current density field at node X,Y,Z */
  virtual void addPxy(double weight[][2][2], int X, int Y, int Z, int ns) = 0;
  /** add an amount of pressure density - direction XZ to current density field at node X,Y,Z */
  virtual void addPxz(double weight[][2][2], int X, int Y, int Z, int ns) = 0;
  /** add an amount of pressure density - direction YY to current density field at node X,Y,Z */
  virtual void addPyy(double weight[][2][2], int X, int Y, int Z, int ns) = 0;
  /** add an amount of pressure density - direction YZ to current density field at node X,Y,Z */
  virtual void addPyz(double weight[][2][2], int X, int Y, int Z, int ns) = 0;
  /** add an amount of pressure density - direction ZZ to current density field at node X,Y,Z */
  virtual void addPzz(double weight[][2][2], int X, int Y, int Z, int ns) = 0;
  /** communicate ghost for grid -> Particles interpolation */
  virtual void communicateGhostP2G(int ns, int bcFaceXright, int bcFaceXleft, int bcFaceYright, int bcFaceYleft, VirtualTopology3D * vct) = 0;
  /** Sum density over different species */
  virtual void sumOverSpecies(VirtualTopology3D * vct) = 0;
  /** Sum current over different species */
  virtual void sumOverSpeciesJ() = 0;
  /** communicate ghost for densities and interp rho from node to center */
  virtual void interpDensitiesN2C(VirtualTopology3D * vct, Grid * grid) = 0;
  /** add a perturbation to charge density */
  virtual void AddPerturbationRho(double deltaBoB, double kx, double ky, double kz, double Bx_mod, double By_mod, double Bz_mod, double ne_mod, double ne_phase, double ni_mod, double ni_phase, double B0, Grid *grid) = 0;
  /** add a perturbattion to the EM field */
  virtual void AddPerturbation(double deltaBoB, double kx, double ky, double kz, double Ex_mod, double Ex_phase, double Ey_mod, double Ey_phase, double Ez_mod, double Ez_phase, double Bx_mod, double Bx_phase, double By_mod, double By_phase, double Bz_mod, double Bz_phase, double B0, Grid *grid) = 0;

    //Ohm's Ohms law things
    // This function: at some particular steps will collect moments
    // Jx_collect, Jy_collect, Jz_collect, rho, PXZ, PYZ
    virtual void sumOverSpeciesOhms(VirtualTopology3D * vct) = 0;
    virtual void setZeroDensitiesOhms() = 0;
    
  // //////////////// FIELD SOLUTION /////////////////////////////////////////
  /** Image of Maxwell solver */
  virtual void MaxwellImage(double *im, double *vector, Grid * grid, VirtualTopology3D * vct) = 0;
  /** maxwell Source */
  virtual void MaxwellSource(double *bkrylov, Grid * grid, VirtualTopology3D * vct) = 0;
  /** Image of Poisson Solver */
  virtual void PoissonImage(double *image, double *vector, Grid * grid, VirtualTopology3D * vct) = 0;

};
#endif
