#ifndef SerialIO_H
#define SerialIO_H

// A corrected version:
// 1) Save full e, full i moments for reconnection simulations [CS els, CS ions, bckgr els, bckgr ions]
// 2) Save smoothed temporary moments (if needed), which are used for Ohm's law terms calculation
// 3) Save Ohm's law terms: convective VxB, div P, (v grad) v

#include <iostream>
#include <sstream>
#include <string>
#include <fstream>


using std::string;
using std::stringstream;
using std::ofstream;
using std::cout;
using std::endl;



/** write to binary file Tracked particles */
inline void writeTrackedParticles(int myrank, int cycle,Particles3D *part){
    // stream file to be opened and managed
    string temp;
    stringstream ss;
    stringstream cc;
    ss << myrank;
    cc << cycle;
    temp = "./data/TrackedParticles_proc" + ss.str() + "_cycle"+ cc.str() ;
    temp += ".bin";
    ofstream my_file(temp.c_str(),ios::binary);
    // first data is NOP = number of particles
    int nop = ((int) part->getNOP());
    my_file.write((char*)&nop,sizeof(nop));
    // then ID, X, Y, Z, U, V, W, Q
    double dvalue;
    int ivalue;
    for (int i=0; i < ((int)part->getNOP()); i++){
        ivalue = ((int) part->getParticleID(i));
        my_file.write((char*)&ivalue,sizeof(ivalue));
        dvalue = part->getX(i);
        my_file.write((char*)&dvalue,sizeof(dvalue));
        dvalue = part->getY(i);
        my_file.write((char*)&dvalue,sizeof(dvalue));
        dvalue = part->getZ(i);
        my_file.write((char*)&dvalue,sizeof(dvalue));
        dvalue = part->getU(i);
        my_file.write((char*)&dvalue,sizeof(dvalue));
        dvalue = part->getV(i);
        my_file.write((char*)&dvalue,sizeof(dvalue));
        dvalue = part->getW(i);
        my_file.write((char*)&dvalue,sizeof(dvalue));
        dvalue = part->getQ(i);
        my_file.write((char*)&dvalue,sizeof(dvalue));
    }
    my_file.close();
}



/** write to an ASCII file Electrostatic Potential */
inline void writePHIascii1D(string filename, int myrank, int cycle, Grid *grid,Field* field){
      // stream file to be opened and managed
      string temp;
      stringstream ss;
      stringstream cc;
      ss << myrank;
      cc << cycle;
      temp = "./data/" + filename +"_proc" + ss.str()  + "_cycle"+ cc.str() ;
      temp += ".txt";
      cout << "Opening file: " << temp << endl;
      ofstream my_file(temp.c_str());
	  int nxc = grid->getNXC();
	 
      for (int i=1; i < grid->getNXC()-1; i++){
        my_file <<  grid->getXC(i,0,0) << "\t " << (field->getPHI())[i][0][0]<< endl;
	  }
      my_file.close();
      
     
}
inline void writeRHOascii1D(string filename, int myrank, int cycle, Grid *grid,Field* field){
      // stream file to be opened and managed
      string temp;
      stringstream ss;
      stringstream cc;
      ss << myrank;
      cc << cycle;
      temp = "./data/" + filename +"_proc" + ss.str() + "_cycle"+ cc.str() ;
      temp += ".txt";
      cout << "Opening file: " << temp << endl;
      ofstream my_file(temp.c_str());
	  int nxc = grid->getNXC();
	 
      for (int i=0; i < grid->getNXN(); i++){
        my_file <<  grid->getXN(i,0,0) << "\t " << field->getRHOn(i,0,0)<< endl;
	  }
      my_file.close();
}
inline void writeRHOascii2D(string filename, int myrank, int cycle, Grid *grid,Field* field){
      // stream file to be opened and managed
      string temp;
      stringstream ss;
      stringstream cc;
      ss << myrank;
      cc << cycle;
      temp = "./data/" + filename +"_proc" + ss.str() + "_cycle"+ cc.str() ;
      temp += ".txt";
      cout << "Opening file: " << temp << endl;
      ofstream my_file(temp.c_str());
	  int nxc = grid->getNXC();
	 
       for (int i=1; i < grid->getNXN()-1; i++){
	    for (int j=1; j < grid->getNYN()-1; j++){
          my_file << field->getRHOn(i,j,0)<< "\t";
	    }
		my_file << endl;
	  }
      my_file.close();
}

inline void writeEyascii2D(string filename, int myrank, int cycle, Grid *grid,Field* field){
    // stream file to be opened and managed
    string temp;
    stringstream ss;
    stringstream cc;
    ss << myrank;
    cc << cycle;
    temp = "./data/" + filename +"_proc" + ss.str() + "_cycle"+ cc.str() ;
    temp += ".txt";
    cout << "Opening file: " << temp << endl;
    ofstream my_file(temp.c_str());
    int nxc = grid->getNXC();
    
    for (int i=1; i < grid->getNXN()-1; i++){
	    for (int j=1; j < grid->getNYN()-1; j++){
            my_file << field->getEy(i,j,0)<< "\t";
	    }
		my_file << endl;
    }
    my_file.close();
}

inline void VTK_Write_Scalars(int cycle, Grid *grid,Field* field){
    // stream file to be opened and managed
    string filename = "rhoe";
    string temp;
    stringstream cc;
    cc << cycle;
    temp = "./data/" + filename + "_"+ cc.str() ;
    temp += ".vtk";
    cout << "Opening file: " << temp << endl;
    
    int nxn = grid->getNXN();
    int nyn = grid->getNYN();
    
    double dx = grid->getDX();
    double dy = grid->getDY();
    
    ofstream my_file(temp.c_str());
    my_file << "# vtk DataFile Version 1.0" << endl;
    my_file << "Electron Density from iPIC" << endl;
    my_file << "ASCII" << endl;
    my_file << "DATASET STRUCTURED_POINTS"  << endl;
    my_file << "DIMENSIONS " << (nxn-2) << " " << (nyn-2) << " " << 1 << endl;
    my_file << "ORIGIN " << 0.0 << " " << 0.0 << " " << 0.0 << endl;
    my_file << "SPACING " << dx << " " << dy << " " << 1.0 << endl;
    my_file << "POINT_DATA " << (nxn-2)*(nyn-2) << endl;
    my_file << "SCALARS rhoe float" << endl;
    my_file << "LOOKUP_TABLE default" << endl;
    
    
	    for (int j=1; j < nyn-1; j++){
            for (int i=1; i < nxn-1; i++){
            my_file << field->getRHOns(i,j,1,0) << endl;
	    }
    }
    my_file.close();
    
    filename = "rhoi";
    temp = "./data/" + filename + "_"+ cc.str() ;
    temp += ".vtk";
    cout << "Opening file: " << temp << endl;
    ofstream my_file2(temp.c_str());
    my_file2 << "# vtk DataFile Version 1.0" << endl;
    my_file2 << "Ion Density from iPIC" << endl;
    my_file2 << "ASCII" << endl;
    my_file2 << "DATASET STRUCTURED_POINTS"  << endl;
    my_file2 << "DIMENSIONS " << (nxn-2) << " " << (nyn-2) << " " << 1 << endl;
    my_file2 << "ORIGIN " << 0.0 << " " << 0.0 << " " << 0.0 << endl;
    my_file2 << "SPACING " << dx << " " << dy << " " << 1.0 << endl;
    my_file2 << "POINT_DATA " << (nxn-2)*(nyn-2) << endl;
    my_file2 << "SCALARS rhoi float" << endl;
    my_file2 << "LOOKUP_TABLE default" << endl;
    
    
	    for (int j=1; j < grid->getNYN()-1; j++){
            for (int i=1; i < grid->getNXN()-1; i++){
            my_file2 << field->getRHOns(i,j,1,1) << endl;
	    }
    }
    my_file2.close();
    
    filename = "rho_net";
    temp = "./data/" + filename + "_"+ cc.str() ;
    temp += ".vtk";
    cout << "Opening file: " << temp << endl;
    ofstream my_file1(temp.c_str());
    my_file1 << "# vtk DataFile Version 1.0" << endl;
    my_file1 << "Net Charge Density from iPIC" << endl;
    my_file1 << "ASCII" << endl;
    my_file1 << "DATASET STRUCTURED_POINTS"  << endl;
    my_file1 << "DIMENSIONS " << (nxn-2) << " " << (nyn-2) << " " << 1 << endl;
    my_file1 << "ORIGIN " << 0.0 << " " << 0.0 << " " << 0.0 << endl;
    my_file1 << "SPACING " << dx << " " << dy << " " << 1.0 << endl;
    my_file1 << "POINT_DATA " << (nxn-2)*(nyn-2) << endl;
    my_file1 << "SCALARS rhonet float" << endl;
    my_file1 << "LOOKUP_TABLE default" << endl;
    
    
	    for (int j=1; j < grid->getNYN()-1; j++){
            for (int i=1; i < grid->getNXN()-1; i++){
            my_file1 << field->getRHOn(i,j,0) << endl;
	    }
    }
    my_file1.close();
    
    
}

inline void VTK_Write_Scalars_combined(int cycle, Grid *grid,Field* field){

    // A. Divin
    // indicate species to be combined into one grid. Not very elegant but works.
    const int combined_e[] = {0, 2};
    const int combined_i[] = {1, 3};
    std::vector<int> combined_i_vector (combined_i, combined_i + sizeof(combined_i) / sizeof(combined_i[0]) );
    std::vector<int> combined_e_vector (combined_e, combined_e + sizeof(combined_e) / sizeof(combined_e[0]) );


    // stream file to be opened and managed
    
    string filename = "rhoe";
    string temp;
    stringstream cc;
    cc << cycle;
    temp = "./data/" + filename + "_"+ cc.str() ;
    temp += ".vtk";
    cout << "Opening file: " << temp << endl;
    
    int nxn = grid->getNXN();
    int nyn = grid->getNYN();
    
    double dx = grid->getDX();
    double dy = grid->getDY();
    
    ofstream my_file(temp.c_str());
    my_file << "# vtk DataFile Version 1.0" << endl;
    my_file << "Electron Density from iPIC" << endl;
    my_file << "ASCII" << endl;
    my_file << "DATASET STRUCTURED_POINTS"  << endl;
    my_file << "DIMENSIONS " << (nxn-2) << " " << (nyn-2) << " " << 1 << endl;
    my_file << "ORIGIN " << 0.0 << " " << 0.0 << " " << 0.0 << endl;
    my_file << "SPACING " << dx << " " << dy << " " << 1.0 << endl;
    my_file << "POINT_DATA " << (nxn-2)*(nyn-2) << endl;
    my_file << "SCALARS rhoe float" << endl;
    my_file << "LOOKUP_TABLE default" << endl;
    
    
    for (int j=1; j < nyn-1; j++){
        for (int i=1; i < nxn-1; i++){
            my_file << field->getRHOns_combined_vector(i,j,1,combined_e_vector,-4*M_PI) << endl;
	    }
    }
    my_file.close();
    
    filename = "rhoi";
    temp = "./data/" + filename + "_"+ cc.str() ;
    temp += ".vtk";
    cout << "Opening file: " << temp << endl;
    ofstream my_file2(temp.c_str());
    my_file2 << "# vtk DataFile Version 1.0" << endl;
    my_file2 << "Ion Density from iPIC" << endl;
    my_file2 << "ASCII" << endl;
    my_file2 << "DATASET STRUCTURED_POINTS"  << endl;
    my_file2 << "DIMENSIONS " << (nxn-2) << " " << (nyn-2) << " " << 1 << endl;
    my_file2 << "ORIGIN " << 0.0 << " " << 0.0 << " " << 0.0 << endl;
    my_file2 << "SPACING " << dx << " " << dy << " " << 1.0 << endl;
    my_file2 << "POINT_DATA " << (nxn-2)*(nyn-2) << endl;
    my_file2 << "SCALARS rhoi float" << endl;
    my_file2 << "LOOKUP_TABLE default" << endl;
    
    
    for (int j=1; j < grid->getNYN()-1; j++){
        for (int i=1; i < grid->getNXN()-1; i++){
            my_file2 << field->getRHOns_combined_vector(i,j,1,combined_i_vector,4*M_PI) << endl;
	    }
    }
    my_file2.close();
    
    filename = "rho_net";
    temp = "./data/" + filename + "_"+ cc.str() ;
    temp += ".vtk";
    cout << "Opening file: " << temp << endl;
    ofstream my_file1(temp.c_str());
    my_file1 << "# vtk DataFile Version 1.0" << endl;
    my_file1 << "Net Charge Density from iPIC" << endl;
    my_file1 << "ASCII" << endl;
    my_file1 << "DATASET STRUCTURED_POINTS"  << endl;
    my_file1 << "DIMENSIONS " << (nxn-2) << " " << (nyn-2) << " " << 1 << endl;
    my_file1 << "ORIGIN " << 0.0 << " " << 0.0 << " " << 0.0 << endl;
    my_file1 << "SPACING " << dx << " " << dy << " " << 1.0 << endl;
    my_file1 << "POINT_DATA " << (nxn-2)*(nyn-2) << endl;
    my_file1 << "SCALARS rhonet float" << endl;
    my_file1 << "LOOKUP_TABLE default" << endl;
    
    
    for (int j=1; j < grid->getNYN()-1; j++){
        for (int i=1; i < grid->getNXN()-1; i++){
            my_file1 << field->getRHOn(i,j,0) << endl;
	    }
    }
    my_file1.close();
    
    
}

inline void VTK_Write_Ohms(int cycle,Grid *grid,Field *field,Collective *col){
    string filename = "Ez_Alfven";
    string temp;
    stringstream cc;
    cc << cycle;
    temp = "./data_Ohm/" + filename + "_"+ cc.str() ;
    temp += ".vtk";
    cout << "Opening file: " << temp << endl;
    
    int nxn = grid->getNXN();
    int nyn = grid->getNYN();
    
    double dx = grid->getDX();
    double dy = grid->getDY();
    double Va;
    double temp_variable1, temp_variable2, temp_variable3, temp_variable4;
    // temp_variable3 is a simple plug for time=0 case, when moments are not collected yet. Passed as an argument from iPIC3D
    if(cycle==0)
        temp_variable3=0.0;
    else
        temp_variable3=1.0;
    
    Va = col->getB0x(); // Works for GEM-like reconnection only. If need general Parsek units, just put Va=1
    
    
    ofstream my_file(temp.c_str());
    my_file << "# vtk DataFile Version 1.0" << endl;
    my_file << "Ez from iPIC" << endl;
    my_file << "ASCII" << endl;
    my_file << "DATASET STRUCTURED_POINTS"  << endl;
    my_file << "DIMENSIONS " << (nxn-2) << " " << (nyn-2) << " " << 1 << endl;
    my_file << "ORIGIN " << 0.0 << " " << 0.0 << " " << 0.0 << endl;
    my_file << "SPACING " << dx << " " << dy << " " << 1.0 << endl;
    my_file << "POINT_DATA " << (nxn-2)*(nyn-2) << endl;
    my_file << "SCALARS Ez float" << endl;
    my_file << "LOOKUP_TABLE default" << endl;
    
    
    for (int j=1; j < nyn-1; j++){
        for (int i=1; i < nxn-1; i++){
            my_file << (field->getEz(i,j,1))/(Va*Va) << endl;
	    }
    }
    my_file.close();
    
    filename = "VeXB";
    temp = "./data_Ohm/" + filename + "_"+ cc.str() ;
    temp += ".vtk";
    cout << "Opening file: " << temp << endl;
    ofstream my_file2(temp.c_str());
    my_file2 << "# vtk DataFile Version 1.0" << endl;
    my_file2 << "Convective electric field from iPIC" << endl;
    my_file2 << "ASCII" << endl;
    my_file2 << "DATASET STRUCTURED_POINTS"  << endl;
    my_file2 << "DIMENSIONS " << (nxn-2) << " " << (nyn-2) << " " << 1 << endl;
    my_file2 << "ORIGIN " << 0.0 << " " << 0.0 << " " << 0.0 << endl;
    my_file2 << "SPACING " << dx << " " << dy << " " << 1.0 << endl;
    my_file2 << "POINT_DATA " << (nxn-2)*(nyn-2) << endl;
    my_file2 << "SCALARS VeXB float" << endl;
    my_file2 << "LOOKUP_TABLE default" << endl;
    // field->getBx(i,j,1), field->getJxs_collect(i,j,1,0) or field->getJxs_collect(i,j,1,1)
    // Econvze=squeeze(vxe02_Alfven_norm).*squeeze(By_Alfven) - squeeze(vye02_Alfven_norm).*squeeze(Bx_Alfven);
    // vxe02_Alfven_norm=squeeze(smooth_3DA(((Jxs0{it}(:,:,:))+(Jxs2{it}(:,:,:)))./(rhos0{it}(:,:,:)+rhos2{it}(:,:,:))/va,smoothx,smoothy,smoothz));

    for (int j=1; j < grid->getNYN()-1; j++){
        for (int i=1; i < grid->getNXN()-1; i++){
        
            if (field->getRHOns_collect(i,j,1,0) == 0.0)
                temp_variable1=0.0;
            else
                temp_variable1= (((field->getJxs_collect(i,j,1,0))/(field->getRHOns_collect(i,j,1,0)))*(field->getBy(i,j,1)) -  ((field->getJys_collect(i,j,1,0))/(field->getRHOns_collect(i,j,1,0)))*(field->getBx(i,j,1)))/(Va*Va);
            my_file2 << temp_variable1*temp_variable3 << endl;
            
	    }
    }
    my_file2.close();


    // Note: qom = -256, Pressures, currents, rhos are all signed (depend on the charge sign). Check carefully when plotting the terms

    filename = "dyPeyz";
    temp = "./data_Ohm/" + filename + "_"+ cc.str() ;
    temp += ".vtk";
    cout << "Opening file: " << temp << endl;
    ofstream my_file3(temp.c_str());
    my_file3 << "# vtk DataFile Version 1.0" << endl;
    my_file3 << "dyPeyz from iPIC" << endl;
    my_file3 << "ASCII" << endl;
    my_file3 << "DATASET STRUCTURED_POINTS"  << endl;
    my_file3 << "DIMENSIONS " << (nxn-2) << " " << (nyn-2) << " " << 1 << endl;
    my_file3 << "ORIGIN " << 0.0 << " " << 0.0 << " " << 0.0 << endl;
    my_file3 << "SPACING " << dx << " " << dy << " " << 1.0 << endl;
    my_file3 << "POINT_DATA " << (nxn-2)*(nyn-2) << endl;
    my_file3 << "SCALARS dyPeyz float" << endl;
    my_file3 << "LOOKUP_TABLE default" << endl;
    // field->getBx(i,j,1), field->getJxs_collect(i,j,1,0) or field->getJxs_collect(i,j,1,1)
    // Econvze=squeeze(vxe02_Alfven_norm).*squeeze(By_Alfven) - squeeze(vye02_Alfven_norm).*squeeze(Bx_Alfven);
    // vxe02_Alfven_norm=squeeze(smooth_3DA(((Jxs0{it}(:,:,:))+(Jxs2{it}(:,:,:)))./(rhos0{it}(:,:,:)+rhos2{it}(:,:,:))/va,smoothx,smoothy,smoothz));
    // pyz_real_Alfven_norm=squeeze(smooth_3DA(( (pyz0{it}+pyz2{it})/(double(qom(1))) -((rhos0{it}+rhos2{it})*(1/double(qom(1))).*vye02_Parsek_norm.*vze02_Parsek_norm))*4*pi/(va*va),smoothx,smoothy,smoothz));
    // field->getpYZsn_collect
    for (int j=1; j < grid->getNYN()-1; j++){
        for (int i=1; i < grid->getNXN()-1; i++){
        
            if ((field->getRHOns_collect(i,j,1,0) == 0.0) || (field->getRHOns_collect(i,j-1,1,0) == 0.0) || (field->getRHOns_collect(i,j+1,1,0) == 0.0))
                temp_variable4=0.0;
            else
            {
            temp_variable1=  ((field->getpYZsn_collect(i,j-1,1,0)) - ((field->getJys_collect(i,j-1,1,0))*((field->getJzs_collect(i,j-1,1,0))/(field->getRHOns_collect(i,j-1,1,0)))))/(Va*Va*(col->getQOM(0)));
            temp_variable2=  ((field->getpYZsn_collect(i,j+1,1,0)) - ((field->getJys_collect(i,j+1,1,0))*((field->getJzs_collect(i,j+1,1,0))/(field->getRHOns_collect(i,j+1,1,0)))))/(Va*Va*(col->getQOM(0)));
            temp_variable4=(temp_variable2-temp_variable1)/(2.0*dy*field->getRHOns_collect(i,j,1,0))*temp_variable3; }
            
            my_file3 <<  temp_variable4 << endl;
            
	    }
    }
    my_file3.close();


    filename = "dxPexz";
    temp = "./data_Ohm/" + filename + "_"+ cc.str() ;
    temp += ".vtk";
    cout << "Opening file: " << temp << endl;
    ofstream my_file4(temp.c_str());
    my_file4 << "# vtk DataFile Version 1.0" << endl;
    my_file4 << "dxPexz from iPIC" << endl;
    my_file4 << "ASCII" << endl;
    my_file4 << "DATASET STRUCTURED_POINTS"  << endl;
    my_file4 << "DIMENSIONS " << (nxn-2) << " " << (nyn-2) << " " << 1 << endl;
    my_file4 << "ORIGIN " << 0.0 << " " << 0.0 << " " << 0.0 << endl;
    my_file4 << "SPACING " << dx << " " << dy << " " << 1.0 << endl;
    my_file4 << "POINT_DATA " << (nxn-2)*(nyn-2) << endl;
    my_file4 << "SCALARS dxPexz float" << endl;
    my_file4 << "LOOKUP_TABLE default" << endl;
    // field->getBx(i,j,1), field->getJxs_collect(i,j,1,0) or field->getJxs_collect(i,j,1,1)
    // Econvze=squeeze(vxe02_Alfven_norm).*squeeze(By_Alfven) - squeeze(vye02_Alfven_norm).*squeeze(Bx_Alfven);
    // vxe02_Alfven_norm=squeeze(smooth_3DA(((Jxs0{it}(:,:,:))+(Jxs2{it}(:,:,:)))./(rhos0{it}(:,:,:)+rhos2{it}(:,:,:))/va,smoothx,smoothy,smoothz));
    // pyz_real_Alfven_norm=squeeze(smooth_3DA(( (pyz0{it}+pyz2{it})/(double(qom(1))) -((rhos0{it}+rhos2{it})*(1/double(qom(1))).*vye02_Parsek_norm.*vze02_Parsek_norm))*4*pi/(va*va),smoothx,smoothy,smoothz));
    // field->getpYZsn_collect
    for (int j=1; j < grid->getNYN()-1; j++){
        for (int i=1; i < grid->getNXN()-1; i++){
            if ((field->getRHOns_collect(i,j,1,0) == 0.0) || (field->getRHOns_collect(i-1,j,1,0) == 0.0) || (field->getRHOns_collect(i+1,j,1,0) == 0.0))
                temp_variable4=0.0;
            else
            {
            temp_variable1= ((field->getpXZsn_collect(i-1,j,1,0)) - ((field->getJxs_collect(i-1,j,1,0))*((field->getJzs_collect(i-1,j,1,0))/(field->getRHOns_collect(i-1,j,1,0)))))/(Va*Va*(col->getQOM(0)));
            temp_variable2= ((field->getpXZsn_collect(i+1,j,1,0)) - ((field->getJxs_collect(i+1,j,1,0))*((field->getJzs_collect(i+1,j,1,0))/(field->getRHOns_collect(i+1,j,1,0)))))/(Va*Va*(col->getQOM(0)));
            temp_variable4=(temp_variable2-temp_variable1)/(2.0*dx*field->getRHOns_collect(i,j,1,0))*temp_variable3;
            }
            
            my_file4 <<  temp_variable4  << endl;
            
	    }
    }
    my_file4.close();


    //VexDxVez
    filename = "VexDxVez";
    temp = "./data_Ohm/" + filename + "_"+ cc.str() ;
    temp += ".vtk";
    cout << "Opening file: " << temp << endl;
    ofstream my_file5(temp.c_str());
    my_file5 << "# vtk DataFile Version 1.0" << endl;
    my_file5 << "VexDxVez from iPIC" << endl;
    my_file5 << "ASCII" << endl;
    my_file5 << "DATASET STRUCTURED_POINTS"  << endl;
    my_file5 << "DIMENSIONS " << (nxn-2) << " " << (nyn-2) << " " << 1 << endl;
    my_file5 << "ORIGIN " << 0.0 << " " << 0.0 << " " << 0.0 << endl;
    my_file5 << "SPACING " << dx << " " << dy << " " << 1.0 << endl;
    my_file5 << "POINT_DATA " << (nxn-2)*(nyn-2) << endl;
    my_file5 << "SCALARS VexDxVez float" << endl;
    my_file5 << "LOOKUP_TABLE default" << endl;
    for (int j=1; j < grid->getNYN()-1; j++){
        for (int i=1; i < grid->getNXN()-1; i++){
        
            if ((field->getRHOns_collect(i,j,1,0) == 0.0) || (field->getRHOns_collect(i-1,j,1,0) == 0.0) || (field->getRHOns_collect(i+1,j,1,0) == 0.0))
                temp_variable4=0.0;
            else
            {
            temp_variable1=(field->getJzs_collect(i-1,j,1,0))/(field->getRHOns_collect(i-1,j,1,0));
            temp_variable2=(field->getJzs_collect(i+1,j,1,0))/(field->getRHOns_collect(i+1,j,1,0));
            temp_variable4=((temp_variable2-temp_variable1)/(2.0*dx*Va*Va*(col->getQOM(0))))*((field->getJxs_collect(i,j,1,0))/(field->getRHOns_collect(i,j,1,0)))*temp_variable3;
            }
            my_file5 << temp_variable4 << endl;
            
	    }
    }
    my_file5.close();

  
  
  
    //VexDxVez
    filename = "VeyDyVez";
    temp = "./data_Ohm/" + filename + "_"+ cc.str() ;
    temp += ".vtk";
    cout << "Opening file: " << temp << endl;
    ofstream my_file6(temp.c_str());
    my_file6 << "# vtk DataFile Version 1.0" << endl;
    my_file6 << "VeyDyVez from iPIC" << endl;
    my_file6 << "ASCII" << endl;
    my_file6 << "DATASET STRUCTURED_POINTS"  << endl;
    my_file6 << "DIMENSIONS " << (nxn-2) << " " << (nyn-2) << " " << 1 << endl;
    my_file6 << "ORIGIN " << 0.0 << " " << 0.0 << " " << 0.0 << endl;
    my_file6 << "SPACING " << dx << " " << dy << " " << 1.0 << endl;
    my_file6 << "POINT_DATA " << (nxn-2)*(nyn-2) << endl;
    my_file6 << "SCALARS VeyDyVez float" << endl;
    my_file6 << "LOOKUP_TABLE default" << endl;
    for (int j=1; j < grid->getNYN()-1; j++){
        for (int i=1; i < grid->getNXN()-1; i++){
            
            if ((field->getRHOns_collect(i,j,1,0) == 0.0) || (field->getRHOns_collect(i,j-1,1,0) == 0.0) || (field->getRHOns_collect(i,j+1,1,0) == 0.0))
                temp_variable4=0.0;
            else
            {
            temp_variable1=(field->getJzs_collect(i,j-1,1,0))/(field->getRHOns_collect(i,j-1,1,0));
            temp_variable2=(field->getJzs_collect(i,j+1,1,0))/(field->getRHOns_collect(i,j+1,1,0));
            temp_variable4=((temp_variable2-temp_variable1)/(2.0*dx*Va*Va*(col->getQOM(0))))*((field->getJys_collect(i,j,1,0))/(field->getRHOns_collect(i,j,1,0)))*temp_variable3;
            }
            my_file6 << temp_variable4 << endl;
            
	    }
    }
    my_file6.close();




    
}




inline void VTK_Write_Smoothed(int cycle, Grid *grid,Field* field){
    // stream file to be opened and managed
    string filename = "rhoe_smoothed";
    string temp;
    stringstream cc;
    cc << cycle;
    temp = "./data/" + filename + "_"+ cc.str() ;
    temp += ".vtk";
    cout << "Opening file: " << temp << endl;
    
    int nxn = grid->getNXN();
    int nyn = grid->getNYN();
    
    double dx = grid->getDX();
    double dy = grid->getDY();
    

    ofstream my_file(temp.c_str());
    my_file << "# vtk DataFile Version 1.0" << endl;
    my_file << "Electron Density smoothed over several past sycles (see code) from iPIC" << endl;
    my_file << "ASCII" << endl;
    my_file << "DATASET STRUCTURED_POINTS"  << endl;
    my_file << "DIMENSIONS " << (nxn-2) << " " << (nyn-2) << " " << 1 << endl;
    my_file << "ORIGIN " << 0.0 << " " << 0.0 << " " << 0.0 << endl;
    my_file << "SPACING " << dx << " " << dy << " " << 1.0 << endl;
    my_file << "POINT_DATA " << (nxn-2)*(nyn-2) << endl;
    my_file << "SCALARS rhoe float" << endl;
    my_file << "LOOKUP_TABLE default" << endl;
    
    
    for (int j=1; j < nyn-1; j++){
        for (int i=1; i < nxn-1; i++){
            my_file << field->getRHOns_collect(i,j,1,0) << endl;
	    }
    }
    my_file.close();
    
    filename = "rhoi_smoothed";
    temp = "./data/" + filename + "_"+ cc.str() ;
    temp += ".vtk";
    cout << "Opening file: " << temp << endl;
    ofstream my_file2(temp.c_str());
    my_file2 << "# vtk DataFile Version 1.0" << endl;
    my_file2 << "Ion Density smoothed over several past sycles (see code) from iPIC" << endl;
    my_file2 << "ASCII" << endl;
    my_file2 << "DATASET STRUCTURED_POINTS"  << endl;
    my_file2 << "DIMENSIONS " << (nxn-2) << " " << (nyn-2) << " " << 1 << endl;
    my_file2 << "ORIGIN " << 0.0 << " " << 0.0 << " " << 0.0 << endl;
    my_file2 << "SPACING " << dx << " " << dy << " " << 1.0 << endl;
    my_file2 << "POINT_DATA " << (nxn-2)*(nyn-2) << endl;
    my_file2 << "SCALARS rhoi float" << endl;
    my_file2 << "LOOKUP_TABLE default" << endl;
    
    
    for (int j=1; j < grid->getNYN()-1; j++){
        for (int i=1; i < grid->getNXN()-1; i++){
            my_file2 << field->getRHOns_collect(i,j,1,1) << endl;
	    }
    }
    my_file2.close();



    
    
    filename = "Je_smoothed";
    temp = "./data/" + filename + "_"+ cc.str() ;
    temp += ".vtk";
    cout << "Opening file: " << temp << endl;
    ofstream my_file1(temp.c_str());
    my_file1 << "# vtk DataFile Version 1.0" << endl;
    my_file1 << "Je from iPIC" << endl;
    my_file1 << "ASCII" << endl;
    my_file1 << "DATASET STRUCTURED_POINTS"  << endl;
    my_file1 << "DIMENSIONS " << (nxn-2) << " " << (nyn-2) << " " << 1 << endl;
    my_file1 << "ORIGIN " << 0.0 << " " << 0.0 << " " << 0.0 << endl;
    my_file1 << "SPACING " << dx << " " << dy << " " << 1.0 << endl;
    my_file1 << "POINT_DATA " << (nxn-2)*(nyn-2) << endl;
    my_file1 << "VECTORS Je_smoothed float" << endl;
    
    double Vx, Vy, Vz;
    for (int j=1; j < nyn-1; j++){
        for (int i=1; i < nxn-1; i++){
            Vx=(field->getJxs_collect(i,j,1,0));
            Vy=(field->getJys_collect(i,j,1,0));
            Vz=(field->getJzs_collect(i,j,1,0));
            my_file1 << Vx << " " << Vy <<  " " << Vz <<  endl;
	    }
    }
    my_file1.close();
    
    
    
    
    filename = "Ji_smoothed";
    temp = "./data/" + filename + "_"+ cc.str() ;
    temp += ".vtk";
    cout << "Opening file: " << temp << endl;
    ofstream my_file4(temp.c_str());
    my_file4 << "# vtk DataFile Version 1.0" << endl;
    my_file4 << "Ion Vel from iPIC" << endl;
    my_file4 << "ASCII" << endl;
    my_file4 << "DATASET STRUCTURED_POINTS"  << endl;
    my_file4 << "DIMENSIONS " << (nxn-2) << " " << (nyn-2) << " " << 1 << endl;
    my_file4 << "ORIGIN " << 0.0 << " " << 0.0 << " " << 0.0 << endl;
    my_file4 << "SPACING " << dx << " " << dy << " " << 1.0 << endl;
    my_file4 << "POINT_DATA " << (nxn-2)*(nyn-2) << endl;
    my_file4 << "VECTORS Ji_smoothed float" << endl;
    
    
    for (int j=1; j < nyn-1; j++){
        for (int i=1; i < nxn-1; i++){
            Vx=(field->getJxs_collect(i,j,1,1));
            Vy=(field->getJys_collect(i,j,1,1));
            Vz=(field->getJzs_collect(i,j,1,1));
            my_file4 << Vx << " " << Vy <<  " " << Vz <<  endl;
	    }
    }
    my_file4.close();
    



}



inline void VTK_Write_Vectors(int cycle, Grid *grid,Field* field){
    // stream file to be opened and managed
    string filename = "E";
    string temp;
    stringstream cc;
    cc << cycle;
    temp = "./data/" + filename + "_"+ cc.str() ;
    temp += ".vtk";
    cout << "Opening file: " << temp << endl;
    
    int nxn = grid->getNXN();
    int nyn = grid->getNYN();
    
    double dx = grid->getDX();
    double dy = grid->getDY();
    
    
    
    ofstream my_fileE(temp.c_str());
    my_fileE << "# vtk DataFile Version 1.0" << endl;
    my_fileE << "E iPIC" << endl;
    my_fileE << "ASCII" << endl;
    my_fileE << "DATASET STRUCTURED_POINTS"  << endl;
    my_fileE << "DIMENSIONS " << (nxn-2) << " " << (nyn-2) << " " << 1 << endl;
    my_fileE << "ORIGIN " << 0.0 << " " << 0.0 << " " << 0.0 << endl;
    my_fileE << "SPACING " << dx << " " << dy << " " << 1.0 << endl;
    my_fileE << "POINT_DATA " << (nxn-2)*(nyn-2) << endl;
    my_fileE << "VECTORS E float" << endl;
    
    double Ex = 0, Ey = 0, Ez = 0;
    for (int j=1; j < nyn-1; j++){
        for (int i=1; i < nxn-1; i++){
            Ex = field->getEx(i,j,1);
            if (fabs(Ex) < 1E-8)
                Ex = 0.0;
            Ey = field->getEy(i,j,1);
            if (fabs(Ey) < 1E-8)
                Ey = 0.0;
            Ez = field->getEz(i,j,1);
            if (fabs(Ez) < 1E-8)
                Ez = 0.0;
            my_fileE << Ex << " " << Ey <<  " " << Ez <<  endl;
	    }
    }
    my_fileE.close();
    
    filename = "B";
    temp = "./data/" + filename + "_"+ cc.str() ;
    temp += ".vtk";
    cout << "Opening file: " << temp << endl;
    ofstream my_file2(temp.c_str());
    my_file2 << "# vtk DataFile Version 1.0" << endl;
    my_file2 << "B iPIC" << endl;
    my_file2 << "ASCII" << endl;
    my_file2 << "DATASET STRUCTURED_POINTS"  << endl;
    my_file2 << "DIMENSIONS " << (nxn-2) << " " << (nyn-2) << " " << 1 << endl;
    my_file2 << "ORIGIN " << 0.0 << " " << 0.0 << " " << 0.0 << endl;
    my_file2 << "SPACING " << dx << " " << dy << " " << 1.0 << endl;
    my_file2 << "POINT_DATA " << (nxn-2)*(nyn-2) << endl;
    my_file2 << "VECTORS B float" << endl;
    
    
    for (int j=1; j < nyn-1; j++){
        for (int i=1; i < nxn-1; i++){
            my_file2 << field->getBx(i,j,1) << " " << field->getBy(i,j,1) <<  " " << field->getBz(i,j,1) <<  endl;
	    }
    }
    my_file2.close();
    
    
    filename = "Je";
    temp = "./data/" + filename + "_"+ cc.str() ;
    temp += ".vtk";
    cout << "Opening file: " << temp << endl;
    ofstream my_file1(temp.c_str());
    my_file1 << "# vtk DataFile Version 1.0" << endl;
    my_file1 << "Je from iPIC" << endl;
    my_file1 << "ASCII" << endl;
    my_file1 << "DATASET STRUCTURED_POINTS"  << endl;
    my_file1 << "DIMENSIONS " << (nxn-2) << " " << (nyn-2) << " " << 1 << endl;
    my_file1 << "ORIGIN " << 0.0 << " " << 0.0 << " " << 0.0 << endl;
    my_file1 << "SPACING " << dx << " " << dy << " " << 1.0 << endl;
    my_file1 << "POINT_DATA " << (nxn-2)*(nyn-2) << endl;
    my_file1 << "VECTORS Je float" << endl;
    
    double Vx, Vy, Vz;
    for (int j=1; j < nyn-1; j++){
        for (int i=1; i < nxn-1; i++){
            Vx=(field->getJxs(i,j,1,0));
            Vy=(field->getJys(i,j,1,0));
            Vz=(field->getJzs(i,j,1,0));
            my_file1 << Vx << " " << Vy <<  " " << Vz <<  endl;
	    }
    }
    my_file1.close();
    
    filename = "Ji";
    temp = "./data/" + filename + "_"+ cc.str() ;
    temp += ".vtk";
    cout << "Opening file: " << temp << endl;
    ofstream my_file3(temp.c_str());
    my_file3 << "# vtk DataFile Version 1.0" << endl;
    my_file3 << "Ion Vel from iPIC" << endl;
    my_file3 << "ASCII" << endl;
    my_file3 << "DATASET STRUCTURED_POINTS"  << endl;
    my_file3 << "DIMENSIONS " << (nxn-2) << " " << (nyn-2) << " " << 1 << endl;
    my_file3 << "ORIGIN " << 0.0 << " " << 0.0 << " " << 0.0 << endl;
    my_file3 << "SPACING " << dx << " " << dy << " " << 1.0 << endl;
    my_file3 << "POINT_DATA " << (nxn-2)*(nyn-2) << endl;
    my_file3 << "VECTORS Ji float" << endl;
    
    
    for (int j=1; j < nyn-1; j++){
        for (int i=1; i < nxn-1; i++){
            Vx=(field->getJxs(i,j,1,1));
            Vy=(field->getJys(i,j,1,1));
            Vz=(field->getJzs(i,j,1,1));
            my_file3 << Vx << " " << Vy <<  " " << Vz <<  endl;
	    }
    }
    my_file3.close();
    
    
}


inline void VTK_Write_Vectors_combined(int cycle, Grid *grid,Field* field){

    // A. Divin
    // indicate species to be combined into one grid. Not very elegant but works.
    const int combined_e[] = {0, 2};
    const int combined_i[] = {1, 3};
    std::vector<int> combined_i_vector (combined_i, combined_i + sizeof(combined_i) / sizeof(combined_i[0]) );
    std::vector<int> combined_e_vector (combined_e, combined_e + sizeof(combined_e) / sizeof(combined_e[0]) );

    // stream file to be opened and managed
    string filename = "E";
    string temp;
    stringstream cc;
    cc << cycle;
    temp = "./data/" + filename + "_"+ cc.str() ;
    temp += ".vtk";
    cout << "Opening file: " << temp << endl;
    
    int nxn = grid->getNXN();
    int nyn = grid->getNYN();
    
    double dx = grid->getDX();
    double dy = grid->getDY();
    
    
    
    ofstream my_fileE(temp.c_str());
    my_fileE << "# vtk DataFile Version 1.0" << endl;
    my_fileE << "E iPIC" << endl;
    my_fileE << "ASCII" << endl;
    my_fileE << "DATASET STRUCTURED_POINTS"  << endl;
    my_fileE << "DIMENSIONS " << (nxn-2) << " " << (nyn-2) << " " << 1 << endl;
    my_fileE << "ORIGIN " << 0.0 << " " << 0.0 << " " << 0.0 << endl;
    my_fileE << "SPACING " << dx << " " << dy << " " << 1.0 << endl;
    my_fileE << "POINT_DATA " << (nxn-2)*(nyn-2) << endl;
    my_fileE << "VECTORS E float" << endl;
    
    double Ex = 0, Ey = 0, Ez = 0;
    for (int j=1; j < nyn-1; j++){
        for (int i=1; i < nxn-1; i++){
            Ex = field->getEx(i,j,1);
            if (fabs(Ex) < 1E-8)
                Ex = 0.0;
            Ey = field->getEy(i,j,1);
            if (fabs(Ey) < 1E-8)
                Ey = 0.0;
            Ez = field->getEz(i,j,1);
            if (fabs(Ez) < 1E-8)
                Ez = 0.0;
            my_fileE << Ex << " " << Ey <<  " " << Ez <<  endl;
	    }
    }
    my_fileE.close();
    
    filename = "B";
    temp = "./data/" + filename + "_"+ cc.str() ;
    temp += ".vtk";
    cout << "Opening file: " << temp << endl;
    ofstream my_file2(temp.c_str());
    my_file2 << "# vtk DataFile Version 1.0" << endl;
    my_file2 << "B iPIC" << endl;
    my_file2 << "ASCII" << endl;
    my_file2 << "DATASET STRUCTURED_POINTS"  << endl;
    my_file2 << "DIMENSIONS " << (nxn-2) << " " << (nyn-2) << " " << 1 << endl;
    my_file2 << "ORIGIN " << 0.0 << " " << 0.0 << " " << 0.0 << endl;
    my_file2 << "SPACING " << dx << " " << dy << " " << 1.0 << endl;
    my_file2 << "POINT_DATA " << (nxn-2)*(nyn-2) << endl;
    my_file2 << "VECTORS B float" << endl;
    
    
    for (int j=1; j < nyn-1; j++){
        for (int i=1; i < nxn-1; i++){
            my_file2 << field->getBx(i,j,1) << " " << field->getBy(i,j,1) <<  " " << field->getBz(i,j,1) <<  endl;
	    }
    }
    my_file2.close();
    
    
    filename = "Je";
    temp = "./data/" + filename + "_"+ cc.str() ;
    temp += ".vtk";
    cout << "Opening file: " << temp << endl;
    ofstream my_file1(temp.c_str());
    my_file1 << "# vtk DataFile Version 1.0" << endl;
    my_file1 << "Je from iPIC" << endl;
    my_file1 << "ASCII" << endl;
    my_file1 << "DATASET STRUCTURED_POINTS"  << endl;
    my_file1 << "DIMENSIONS " << (nxn-2) << " " << (nyn-2) << " " << 1 << endl;
    my_file1 << "ORIGIN " << 0.0 << " " << 0.0 << " " << 0.0 << endl;
    my_file1 << "SPACING " << dx << " " << dy << " " << 1.0 << endl;
    my_file1 << "POINT_DATA " << (nxn-2)*(nyn-2) << endl;
    my_file1 << "VECTORS Je float" << endl;
    
    double Vx, Vy, Vz;
    for (int j=1; j < nyn-1; j++){
        for (int i=1; i < nxn-1; i++){
            Vx=(field->getJxs_combined_vector(i,j,1,combined_e_vector,4.0*M_PI));
            Vy=(field->getJys_combined_vector(i,j,1,combined_e_vector,4.0*M_PI));
            Vz=(field->getJzs_combined_vector(i,j,1,combined_e_vector,4.0*M_PI));
            my_file1 << Vx << " " << Vy <<  " " << Vz <<  endl;
	    }
    }
    my_file1.close();
    
    filename = "Ji";
    temp = "./data/" + filename + "_"+ cc.str() ;
    temp += ".vtk";
    cout << "Opening file: " << temp << endl;
    ofstream my_file3(temp.c_str());
    my_file3 << "# vtk DataFile Version 1.0" << endl;
    my_file3 << "Ion Vel from iPIC" << endl;
    my_file3 << "ASCII" << endl;
    my_file3 << "DATASET STRUCTURED_POINTS"  << endl;
    my_file3 << "DIMENSIONS " << (nxn-2) << " " << (nyn-2) << " " << 1 << endl;
    my_file3 << "ORIGIN " << 0.0 << " " << 0.0 << " " << 0.0 << endl;
    my_file3 << "SPACING " << dx << " " << dy << " " << 1.0 << endl;
    my_file3 << "POINT_DATA " << (nxn-2)*(nyn-2) << endl;
    my_file3 << "VECTORS Ji float" << endl;
    
    
    for (int j=1; j < nyn-1; j++){
        for (int i=1; i < nxn-1; i++){
            Vx=(field->getJxs_combined_vector(i,j,1,combined_i_vector,4.0*M_PI));
            Vy=(field->getJys_combined_vector(i,j,1,combined_i_vector,4.0*M_PI));
            Vz=(field->getJzs_combined_vector(i,j,1,combined_i_vector,4.0*M_PI));
            my_file3 << Vx << " " << Vy <<  " " << Vz <<  endl;
	    }
    }
    my_file3.close();
    
    
}



inline void writeRHOascii2Del(string filename, int myrank, int cycle, Grid *grid,Field* field){
      // stream file to be opened and managed
      string temp;
      stringstream ss;
      stringstream cc;
      ss << myrank;
      cc << cycle;
      temp = "./data/" + filename +"_proc" + ss.str() + "_cycle"+ cc.str() ;
      temp += ".txt";
      cout << "Opening file: " << temp << endl;
      ofstream my_file(temp.c_str());
	  int nxc = grid->getNXC();
	 
      for (int i=1; i < grid->getNXN()-1; i++){
	    for (int j=1; j < grid->getNYN()-1; j++){
          my_file << field->getRHOns(i,j,0,0)<< "\t";
	    }
		my_file << endl;
	  }
      my_file.close();
}
inline void writeVyascii2Del(string filename, int myrank, int cycle, Grid *grid,Field* field){
    // stream file to be opened and managed
    string temp;
    stringstream ss;
    stringstream cc;
    ss << myrank;
    cc << cycle;
    temp = "./data/" + filename +"_proc" + ss.str() + "_cycle"+ cc.str() ;
    temp += ".txt";
    cout << "Opening file: " << temp << endl;
    ofstream my_file(temp.c_str());
    int nxc = grid->getNXC();
    double v = 0.0;
    for (int i=1; i < grid->getNXN()-1; i++){
	    for (int j=1; j < grid->getNYN()-1; j++){
            if (field->getRHOns(i,j,0,0)==0)
                v = 0.0;
            else
                v = ((field->getJys(i,j,0,0))/(field->getRHOns(i,j,0,0)));
            my_file << v << "\t";
	    }
		my_file << endl;
    }
    my_file.close();
}
inline void writeVxascii2Del(string filename, int myrank, int cycle, Grid *grid,Field* field){
    // stream file to be opened and managed
    string temp;
    stringstream ss;
    stringstream cc;
    ss << myrank;
    cc << cycle;
    temp = "./data/" + filename +"_proc" + ss.str() + "_cycle"+ cc.str() ;
    temp += ".txt";
    cout << "Opening file: " << temp << endl;
    ofstream my_file(temp.c_str());
    int nxc = grid->getNXC();
    
    double v;
    
    for (int i=1; i < grid->getNXN()-1; i++){
	    for (int j=1; j < grid->getNYN()-1; j++){
            if (field->getRHOns(i,j,0,0)==0)
                v = 0.0;
            else
                v = ((field->getJxs(i,j,0,0))/(field->getRHOns(i,j,0,0)));
            my_file << v << "\t";
	    }
		my_file << endl;
    }
    my_file.close();
}
inline void writeVxascii2Dion(string filename, int myrank, int cycle, Grid *grid,Field* field){
    // stream file to be opened and managed
    string temp;
    stringstream ss;
    stringstream cc;
    ss << myrank;
    cc << cycle;
    temp = "./data/" + filename +"_proc" + ss.str() + "_cycle"+ cc.str() ;
    temp += ".txt";
    cout << "Opening file: " << temp << endl;
    ofstream my_file(temp.c_str());
    int nxc = grid->getNXC();
    double v;
    
    for (int i=1; i < grid->getNXN()-1; i++){
	    for (int j=1; j < grid->getNYN()-1; j++){
            if (field->getRHOns(i,j,0,1)==0)
                v = 0.0;
            else
                v = ((field->getJxs(i,j,0,1))/(field->getRHOns(i,j,0,1)));
            my_file << v << "\t";
	    }
		my_file << endl;
    }
    my_file.close();
}
inline void writeVyascii2Dion(string filename, int myrank, int cycle, Grid *grid,Field* field){
    // stream file to be opened and managed
    string temp;
    stringstream ss;
    stringstream cc;
    ss << myrank;
    cc << cycle;
    temp = "./data/" + filename +"_proc" + ss.str() + "_cycle"+ cc.str() ;
    temp += ".txt";
    cout << "Opening file: " << temp << endl;
    ofstream my_file(temp.c_str());
    int nxc = grid->getNXC();
    double v = 0;
    for (int i=1; i < grid->getNXN()-1; i++){
	    for (int j=1; j < grid->getNYN()-1; j++){
            if (field->getRHOns(i,j,0,1)==0)
                v = 0.0;
            else
                v = ((field->getJys(i,j,0,1))/(field->getRHOns(i,j,0,1)));
            my_file << v << "\t";
	    }
		my_file << endl;
    }
    my_file.close();
}
inline void writeRHOascii2Dion(string filename, int myrank, int cycle, Grid *grid,Field* field){
      // stream file to be opened and managed
      string temp;
      stringstream ss;
      stringstream cc;
      ss << myrank;
      cc << cycle;
      temp = "./data/" + filename +"_proc" + ss.str() + "_cycle"+ cc.str() ;
      temp += ".txt";
      cout << "Opening file: " << temp << endl;
      ofstream my_file(temp.c_str());
	  int nxc = grid->getNXC();
	 
       for (int i=1; i < grid->getNXN()-1; i++){
	    for (int j=1; j < grid->getNYN()-1; j++){
          my_file << field->getRHOns(i,j,0,1)<< "\t";
	    }
		my_file << endl;
	  }
      my_file.close();
}
/** write an ASCII with Ex */
inline void writeExascii1D(string filename, int myrank, int cycle, Grid *grid,Field* field){
      // stream file to be opened and managed
      string temp;
      stringstream ss;
      stringstream cc;
      ss << myrank;
      cc << cycle;
      temp = "./data/" + filename + "_cycle"+ cc.str() +"_proc" + ss.str();
      temp += ".txt";
      cout << "Opening file: " << temp << endl;
      ofstream my_file(temp.c_str());
      for (int i=0; i < (grid->getNXN()); i++){
         my_file <<  field->getRHOns(i,0,0,0)  << endl;
      }
      my_file.close();
}
/** write an ASCII with the heat flux */
inline void writeJx1D(string filename, int myrank, int cycle, Grid *grid,Field* field){
      // stream file to be opened and managed
      string temp;
      stringstream ss;
      stringstream cc;
      ss << myrank;
      cc << cycle;
      temp = "./data/" + filename +"_proc" + ss.str() + "_cycle"+ cc.str() ;
      temp += ".txt";
      cout << "Opening file: " << temp << endl;
      ofstream my_file(temp.c_str());
      for (int i=0; i < (grid->getNXN()); i++){
         my_file <<  field->getJx(i,0,0)  << endl;
      }
      my_file.close();
}
/** write an ASCII with the heat flux */
inline void writeHeatFluxascii1D(string filename, int myrank, int cycle, Grid *grid,Field* field){
      // stream file to be opened and managed
      string temp;
      stringstream ss;
      stringstream cc;
      ss << myrank;
      cc << cycle;
      temp = "./data/" + filename + "_proc" + ss.str() +  "_cycle"+ cc.str() ;
      temp += ".txt";
      cout << "Opening file: " << temp << endl;
      ofstream my_file(temp.c_str());
      for (int i=0; i < (grid->getNXN()); i++){
         my_file <<  field->getJy(i,0,0)  << endl;
      }
      my_file.close();
}
/** write to an ASCII file  particles position */
inline void writeParticles1D(string filename, int myrank, int cycle,Particles *part){
      // stream file to be opened and managed
      string temp;
      stringstream ss;
      stringstream cc;
      ss << myrank;
      cc << cycle;
      temp = "./data/" + filename +"_proc" + ss.str() + "_cycle"+ cc.str() ;
      temp += ".txt";
      cout << "Opening file: " << temp << endl;
      ofstream my_file(temp.c_str());
      for (int i=0; i < part->getNOP(); i++){
        my_file << part->getY(i) << "\t" << part->getV(i) << endl;
      }
      my_file.close();
}
/** write to an ASCII file  particles position */
inline void writeParticles1D(string filename, int myrank, int cycle,Particles *part,int p){
      // stream file to be opened and managed
      string temp;
      stringstream ss;
      stringstream cc;
      ss << myrank;
      cc << cycle;
      temp = "./data/" + filename + "_proc" + ss.str() + "_cycle"+ cc.str() ;
      temp += ".txt";
      cout << "Opening file: " << temp << endl;
      ofstream my_file(temp.c_str());
      for (int i=0; i < part->getNOP(); i=i+p){
           my_file << part->getY(i) << "\t" << part->getV(i) << endl;
      }
      my_file.close();
}



#endif
